"""Constants that specify sample sources."""

from enum import StrEnum

METAGENOME_SOURCES = (
    "blood",
    "skin",
    "respiratory",
    "urine",
    "CSF",
    "faeces",
    "environmental",
    "unknown",
    "other",
)

ANALYSIS_SOURCES = (
    "blood",
    "buccal swab",
    "cell-free DNA",
    "cell line",
    "cytology (FFPE)",
    "cytology (not fixed/fresh)",
    "fibroblast",
    "muscle",
    "nail",
    "saliva",
    "skin",
    "tissue (FFPE)",
    "tissue (fresh frozen)",
    "bone marrow",
    "other",
)


class SourceType(StrEnum):
    BLOOD: str = "blood"
    BONE_MARROW: str = "bone marrow"
    BUCCAL_SWAB: str = "buccal swab"
    CELL_FREE_DNA: str = "cell-free DNA"
    CELL_LINE: str = "cell line"
    CYTOLOGY: str = "cytology (not fixed/fresh)"
    CYTOLOGY_FFPE: str = "cytology (FFPE)"
    FFPE: str = "FFPE"
    FIBROBLAST: str = "fibroblast"
    MUSCLE: str = "muscle"
    NAIL: str = "nail"
    OTHER: str = "other"
    SALIVA: str = "saliva"
    SKIN: str = "skin"
    TISSUE: str = "tissue (fresh frozen)"
    TISSUE_FFPE: str = "tissue (FFPE)"
    UNKNOWN: str = "unknown"



"""Test MIP-DNA observations API."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockFixture

from cg.constants.sequencing import SequencingMethod
from cg.exc import LoqusdbDuplicateRecordError
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.models.observations.input_files import MipDNAObservationsInputFiles
from cg.store.models import Case


def test_is_sample_type_eligible_for_observations_upload(
    case_id: str, mip_dna_observations_api: MipDNAObservationsAPI
):
    """Test if the sample type is eligible for observation uploads."""

    # GIVEN a case without tumor samples and a MIP-DNA observations API
    case: Case = mip_dna_observations_api.store.get_case_by_internal_id(case_id)

    # WHEN checking sample type eligibility for a case
    is_sample_type_eligible_for_observations_upload: bool = (
        mip_dna_observations_api.is_sample_type_eligible_for_observations_upload(case)
    )

    # THEN the analysis type should be eligible for observation uploads
    assert is_sample_type_eligible_for_observations_upload


def test_is_sample_type_not_eligible_for_observations_upload(
    case_id: str, mip_dna_observations_api: MipDNAObservationsAPI
):
    """Test if the sample type is not eligible for observation uploads."""

    # GIVEN a case with tumor samples and a MIP-DNA observations API
    case: Case = mip_dna_observations_api.store.get_case_by_internal_id(case_id)
    case.samples[0].is_tumour = True

    # WHEN checking sample type eligibility for a case
    is_sample_type_eligible_for_observations_upload: bool = (
        mip_dna_observations_api.is_sample_type_eligible_for_observations_upload(case)
    )

    # THEN the analysis type should not be eligible for observation uploads
    assert not is_sample_type_eligible_for_observations_upload


def test_is_case_eligible_for_observations_upload(
    case_id: str, mip_dna_observations_api: MipDNAObservationsAPI, mocker: MockFixture
):
    """Test whether a case is eligible for MIP-DNA observation uploads."""

    # GIVEN a case and a MIP-DNA observations API
    case: Case = mip_dna_observations_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a MIP-DNA scenario for Loqusdb uploads
    mocker.patch.object(
        MipDNAAnalysisAPI, "get_data_analysis_type", return_value=SequencingMethod.WGS
    )

    # WHEN checking the upload eligibility for a case
    is_case_eligible_for_observations_upload: bool = (
        mip_dna_observations_api.is_case_eligible_for_observations_upload(case)
    )

    # THEN the case should be eligible for observation uploads
    assert is_case_eligible_for_observations_upload


def test_is_case_not_eligible_for_observations_upload(
    case_id: str, mip_dna_observations_api: MipDNAObservationsAPI, mocker: MockFixture
):
    """Test whether a case is not eligible for MIP-DNA observation uploads."""

    # GIVEN a case and a MIP-DNA observations API
    case: Case = mip_dna_observations_api.analysis_api.status_db.get_case_by_internal_id(case_id)

    # GIVEN a MIP-DNA scenario for Loqusdb uploads with an invalid sequencing method
    mocker.patch.object(
        MipDNAAnalysisAPI, "get_data_analysis_type", return_value=SequencingMethod.WTS
    )

    # WHEN checking the upload eligibility for a case
    is_case_eligible_for_observations_upload: bool = (
        mip_dna_observations_api.is_case_eligible_for_observations_upload(case)
    )

    # THEN the case should not be eligible for observation uploads
    assert not is_case_eligible_for_observations_upload


def test_load_observations(
    case_id: str,
    loqusdb_id: str,
    number_of_loaded_variants: int,
    mip_dna_observations_api: MipDNAObservationsAPI,
    mip_dna_observations_input_files: MipDNAObservationsInputFiles,
    caplog: LogCaptureFixture,
    mocker: MockFixture,
):
    """Test loading of case observations for MIP-DNA."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a MIP-DNA observations API and a list of observations input files

    # GIVEN a MIP-DNA case and a mocked scenario for uploads
    case: Case = mip_dna_observations_api.store.get_case_by_internal_id(case_id)
    mocker.patch.object(
        MipDNAAnalysisAPI, "get_data_analysis_type", return_value=SequencingMethod.WGS
    )
    mocker.patch.object(
        MipDNAObservationsAPI,
        "get_observations_input_files",
        return_value=mip_dna_observations_input_files,
    )
    mocker.patch.object(MipDNAObservationsAPI, "is_duplicate", return_value=False)

    # WHEN loading the case to Loqusdb
    mip_dna_observations_api.load_observations(case)

    # THEN the observations should be loaded without any errors
    assert f"Uploaded {number_of_loaded_variants} variants to Loqusdb" in caplog.text


def test_load_duplicated_observations(
    case_id: str,
    loqusdb_id: str,
    number_of_loaded_variants: int,
    mip_dna_observations_api: MipDNAObservationsAPI,
    mip_dna_observations_input_files: MipDNAObservationsInputFiles,
    caplog: LogCaptureFixture,
    mocker: MockFixture,
):
    """Test loading of duplicated observations for MIP-DNA."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a MIP-DNA observations API and a list of observations input files

    # GIVEN a MIP-DNA case and a mocked scenario for uploads with a case already uploaded
    case: Case = mip_dna_observations_api.store.get_case_by_internal_id(case_id)
    mocker.patch.object(
        MipDNAAnalysisAPI, "get_data_analysis_type", return_value=SequencingMethod.WGS
    )
    mocker.patch.object(
        MipDNAObservationsAPI,
        "get_observations_input_files",
        return_value=mip_dna_observations_input_files,
    )
    mocker.patch.object(MipDNAObservationsAPI, "is_duplicate", return_value=True)

    # WHEN loading the case to Loqusdb
    with pytest.raises(LoqusdbDuplicateRecordError):
        mip_dna_observations_api.load_observations(case)

    # THEN the observations upload should be aborted
    assert f"Case {case_id} has already been uploaded to Loqusdb" in caplog.text



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import (
    get_fastq_delivery_path,
    get_scout_link,
)
from cg.store.models import Case


def get_case_message(case: Case) -> str:
    scout_link: str = get_scout_link(case)
    delivery_path: str = get_fastq_delivery_path(case)
    return (
        f"Hello,\n\n"
        f"The analysis has been uploaded to Scout for the following case:\n\n"
        f"{scout_link}\n\n"
        f"The fastq and analysis files are currently being uploaded to your inbox on Caesar:\n\n"
        f"{delivery_path}"
    )


def get_cases_message(cases: list[Case]) -> str:
    scout_links: list[str] = [get_scout_link(case) for case in cases]
    scout_links_row_separated: str = "\n".join(scout_links)
    delivery_path: str = get_fastq_delivery_path(cases[0])
    return (
        f"Hello,\n\n"
        f"The analyses have been uploaded to Scout for the following cases:\n\n"
        f"{scout_links_row_separated}\n\n"
        f"The fastq and analysis files are currently being uploaded to your inbox on Caesar:\n\n"
        f"{delivery_path}"
    )


class FastqAnalysisScoutMessage(DeliveryMessage):
    def create_message(self, cases: list[Case]) -> str:
        if len(cases) == 1:
            return get_case_message(cases[0])
        return get_cases_message(cases)



from pathlib import Path

import pytest

from cg.models.demultiplex.run_parameters import (
    RunParametersHiSeq,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)


@pytest.fixture(scope="function")
def run_parameters_hiseq_different_index(run_parameters_dir: Path) -> RunParametersHiSeq:
    """Return a HiSeq RunParameters object with different index cycles."""
    path = Path(run_parameters_dir, "RunParameters_hiseq_2500_different_index_cycles.xml")
    return RunParametersHiSeq(run_parameters_path=path)


@pytest.fixture(scope="function")
def run_parameters_novaseq_6000_different_index(
    run_parameters_dir: Path,
) -> RunParametersNovaSeq6000:
    """Return a NovaSeq6000 RunParameters object with different index cycles."""
    path = Path(run_parameters_dir, "RunParameters_novaseq_6000_different_index_cycles.xml")
    return RunParametersNovaSeq6000(run_parameters_path=path)


@pytest.fixture(scope="function")
def run_parameters_novaseq_x_different_index(
    run_parameters_dir: Path,
) -> RunParametersNovaSeqX:
    """Return a NovaSeqX RunParameters object with different index cycles."""
    path = Path(run_parameters_dir, "RunParameters_novaseq_X_different_index_cycles.xml")
    return RunParametersNovaSeqX(run_parameters_path=path)


@pytest.fixture(scope="session")
def hiseq_x_single_index_run_parameters(
    hiseq_x_single_index_run_parameters_path: Path,
) -> RunParametersHiSeq:
    """Return a HiSeqX run parameters object with single index."""
    return RunParametersHiSeq(run_parameters_path=hiseq_x_single_index_run_parameters_path)


@pytest.fixture(scope="session")
def hiseq_x_dual_index_run_parameters(
    hiseq_x_dual_index_run_parameters_path: Path,
) -> RunParametersHiSeq:
    """Return a HiSeqX run parameters object with dual index."""
    return RunParametersHiSeq(run_parameters_path=hiseq_x_dual_index_run_parameters_path)


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_run_parameters(
    hiseq_2500_dual_index_run_parameters_path: Path,
) -> RunParametersHiSeq:
    """Return a HiSeq2500 run parameters object with dual index."""
    return RunParametersHiSeq(run_parameters_path=hiseq_2500_dual_index_run_parameters_path)


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_run_parameters(
    hiseq_2500_custom_index_run_parameters_path: Path,
) -> RunParametersHiSeq:
    """Return a HiSeq2500 run parameters object with custom index."""
    return RunParametersHiSeq(run_parameters_path=hiseq_2500_custom_index_run_parameters_path)


@pytest.fixture
def novaseq_6000_run_parameters_pre_1_5_kits(
    novaseq_6000_run_parameters_pre_1_5_kits_path: Path,
) -> RunParametersNovaSeq6000:
    """Return a NovaSeq6000 run parameters pre 1.5 kit object."""
    return RunParametersNovaSeq6000(
        run_parameters_path=novaseq_6000_run_parameters_pre_1_5_kits_path
    )


@pytest.fixture
def novaseq_6000_run_parameters_post_1_5_kits(novaseq_6000_run_parameters_post_1_5_kits_path: Path):
    """Return a NovaSeq6000 run parameters post 1.5 kit object."""
    return RunParametersNovaSeq6000(
        run_parameters_path=novaseq_6000_run_parameters_post_1_5_kits_path
    )


@pytest.fixture
def novaseq_x_run_parameters(
    novaseq_x_run_parameters_path: Path,
) -> RunParametersNovaSeqX:
    """Return a NovaSeqX run parameters object."""
    return RunParametersNovaSeqX(run_parameters_path=novaseq_x_run_parameters_path)



"""Test for the clean flow cells cmd."""

import logging
from pathlib import Path
from unittest import mock

from click.testing import CliRunner

from cg.cli.clean import clean_flow_cells
from cg.models.cg_config import CGConfig


def test_clean_flow_cells_cmd(
    cli_runner: CliRunner,
    clean_flow_cells_context: CGConfig,
    tmp_flow_cell_to_clean_path: Path,
    tmp_flow_cell_not_to_clean_path: Path,
    caplog,
):
    """Test the clean flow cells command."""
    # GIVEN a config with StatusDB and Housekeeper
    caplog.set_level(logging.DEBUG)
    # GIVEN one flow cell that should be cleaned and one that should not
    assert tmp_flow_cell_not_to_clean_path.exists()
    assert tmp_flow_cell_to_clean_path.exists()

    # WHEN running the clean flow cells cli command
    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        result = cli_runner.invoke(clean_flow_cells, obj=clean_flow_cells_context)

    # THEN assert it exits with success
    assert result.exit_code == 0

    # THEN the flow cell fulfilling all cleaning criteria is deleted
    assert not tmp_flow_cell_to_clean_path.exists()
    assert "Successfully removed the directory and its contents" in caplog.text

    # THEN the flow cell not fulfilling all cleaning criteria is not deleted
    assert tmp_flow_cell_not_to_clean_path.exists()
    assert "CleanFlowCellFailedError" in caplog.text


def test_clean_flow_cells_cmd_dry_run(
    cli_runner: CliRunner,
    clean_flow_cells_context: CGConfig,
    tmp_flow_cell_to_clean_path: Path,
    tmp_flow_cell_not_to_clean_path: Path,
    caplog,
):
    """Test the clean flow cells command using dry-run."""
    # GIVEN a config with StatusDB and Housekeeper
    caplog.set_level(logging.DEBUG)
    # GIVEN one flow cell that should be cleaned and one that should not
    assert tmp_flow_cell_not_to_clean_path.exists()
    assert tmp_flow_cell_to_clean_path.exists()

    # WHEN running the clean flow cells cli command
    with mock.patch(
        "cg.meta.clean.clean_flow_cells.CleanFlowCellAPI.is_directory_older_than_21_days",
        return_value=True,
    ):
        result = cli_runner.invoke(clean_flow_cells, ["--dry-run"], obj=clean_flow_cells_context)

    # THEN assert it exits with success
    assert result.exit_code == 0

    # THEN the directory to clean is not deleted
    assert tmp_flow_cell_to_clean_path.exists()
    assert tmp_flow_cell_not_to_clean_path.exists()
    assert "Would have removed" in caplog.text



"""Fixtures for the sequencing metrics parser tests."""

from pathlib import Path

import pytest

from cg.constants.constants import SCALE_TO_READ_PAIRS
from cg.constants.metrics import DemuxMetricsColumnNames, QualityMetricsColumnNames
from cg.services.bcl_convert_metrics_service.bcl_convert_metrics_service import (
    BCLConvertMetricsService,
)
from cg.services.bcl_convert_metrics_service.models import DemuxMetrics, SequencingQualityMetrics
from cg.services.bcl_convert_metrics_service.parser import MetricsParser


@pytest.fixture(scope="session")
def bcl_convert_metrics_dir_path() -> Path:
    """Return a path to a BCLConvert metrics directory."""
    return Path(
        "tests",
        "fixtures",
        "services",
        "bcl_convert_metrics_service",
        "230622_A00621_0864_AHY7FFDRX2",
    )


@pytest.fixture(scope="session")
def test_sample_internal_id() -> str:
    """Return a test sample internal id."""
    return "ACC11927A2"


@pytest.fixture(scope="session")
def test_lane() -> int:
    """Return a test lane."""
    return 1


@pytest.fixture(scope="session")
def bcl_convert_reads_for_test_sample() -> int:
    """Return the number of reads for the test sample."""
    return 15962796


@pytest.fixture(scope="session")
def bcl_convert_test_q30_bases_percent() -> float:
    """Return the Q30 bases percent for the test sample."""
    return 94


@pytest.fixture(scope="session")
def bcl_convert_test_mean_quality_score_per_lane() -> float:
    """Return the mean quality score for the test sample."""
    return 36.02


@pytest.fixture(scope="session")
def bcl_convert_demux_metric_model_with_data(
    test_lane,
    test_sample_internal_id,
) -> DemuxMetrics:
    """Return a BclConvertDemuxMetrics model with data."""
    return DemuxMetrics(
        **{
            DemuxMetricsColumnNames.LANE.value: test_lane,
            DemuxMetricsColumnNames.SAMPLE_INTERNAL_ID.value: test_sample_internal_id,
            DemuxMetricsColumnNames.READ_PAIR_COUNT.value: 15962796,
        }
    )


@pytest.fixture(scope="session")
def bcl_convert_quality_metric_model_with_data(
    test_lane, test_sample_internal_id
) -> SequencingQualityMetrics:
    """Return a BclConvertQualityMetrics model with data."""
    return SequencingQualityMetrics(
        **{
            QualityMetricsColumnNames.LANE.value: test_lane,
            QualityMetricsColumnNames.SAMPLE_INTERNAL_ID.value: test_sample_internal_id,
            QualityMetricsColumnNames.MEAN_QUALITY_SCORE_Q30.value: 36.15,
            QualityMetricsColumnNames.Q30_BASES_PERCENT.value: 0.95,
            QualityMetricsColumnNames.QUALITY_SCORE_SUM.value: 15004333259,
            QualityMetricsColumnNames.YIELD.value: 415032696,
            QualityMetricsColumnNames.YIELD_Q30.value: 393745856,
        }
    )


@pytest.fixture(scope="session")
def parsed_bcl_convert_metrics(bcl_convert_metrics_dir_path) -> MetricsParser:
    """Return an object with parsed BCLConvert metrics."""
    return MetricsParser(bcl_convert_metrics_dir_path=bcl_convert_metrics_dir_path)


@pytest.fixture(scope="session")
def expected_aggegrated_reads() -> int:
    """Return the expected aggregated reads for metrics file."""
    return 1458658700


@pytest.fixture(scope="session")
def expected_aggregated_undetermined_reads() -> int:
    """Return the expected aggregated undetermined reads for metrics file."""
    return 154578798


@pytest.fixture(scope="session")
def expected_aggregated_percent_q30() -> float:
    """Return the expected aggregated percent Q30 for metrics file."""
    return 0.92


@pytest.fixture(scope="session")
def expected_aggregated_quality_score() -> float:
    """Return the expected aggregated quality score for metrics file."""
    return 35.6


@pytest.fixture(scope="session")
def expected_aggegrated_yield() -> int:
    """Return the expected aggregated yield for metrics file."""
    return 84610417748


@pytest.fixture(scope="session")
def expected_aggegrated_yield_q30() -> int:
    """Return the expected aggregated yield Q30 for metrics file."""
    return 78046839027



"""Test methods for cg/cli/set/samples"""

import logging

import pytest

from cg.cli.set.base import samples
from cg.models.cg_config import CGConfig
from cg.store.store import Store

SUCCESS = 0


@pytest.mark.parametrize("identifier_key", ["name", "internal_id"])
def test_set_samples_by_identifiers(
    cli_runner, base_context, base_store: Store, identifier_key, helpers, caplog
):
    # GIVEN a database with a sample
    sample_obj = helpers.add_sample(base_store)
    identifier_value = getattr(sample_obj, identifier_key)

    # WHEN calling set samples with valid identifiers
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples,
            ["-id", identifier_key, identifier_value, "-y", "--skip-lims"],
            obj=base_context,
        )

    # THEN it should name the sample to be changed
    assert result.exit_code == SUCCESS
    assert sample_obj.internal_id in caplog.text
    assert sample_obj.name in caplog.text


def test_set_samples_by_invalid_identifier(
    cli_runner, base_context, base_store: Store, helpers, caplog
):
    # GIVEN a database with a sample that belongs to a case
    sample_obj = helpers.add_sample(base_store)

    # WHEN calling set samples with an identifier not existing on sample
    bad_identifier = "bad_identifier"
    assert not hasattr(sample_obj, bad_identifier)
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples, ["-id", bad_identifier, "any_value", "-y", "--skip-lims"], obj=base_context
        )

    # THEN it should fail and not name the sample to be changed
    assert sample_obj.internal_id not in caplog.text
    assert sample_obj.name not in caplog.text
    assert result.exit_code != SUCCESS


def test_set_samples_by_case_id(cli_runner, base_context, base_store: Store, helpers, caplog):
    # GIVEN a database with a sample that belongs to a case
    case_obj = helpers.add_case(store=base_store)
    sample_obj = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_obj)

    # WHEN calling set samples with case_id
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples, [case_obj.internal_id, "-y", "--skip-lims"], obj=base_context
        )

    # THEN it should name the sample to be changed
    assert result.exit_code == SUCCESS
    assert sample_obj.internal_id in caplog.text
    assert sample_obj.name in caplog.text


def test_set_samples_by_invalid_case_id(
    cli_runner, base_context, base_store: Store, helpers, caplog
):
    # GIVEN a database with a sample that belongs to a case
    sample_obj = helpers.add_sample(base_store)

    # WHEN calling set samples with an identifier not existing on sample
    non_existing_case = "not_a_case"
    assert not base_store.get_case_by_internal_id(internal_id=non_existing_case)
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples, [non_existing_case, "-y", "--skip-lims"], obj=base_context
        )

    # THEN it should fail and not name the sample to be changed
    assert result.exit_code != SUCCESS
    assert sample_obj.internal_id not in caplog.text
    assert sample_obj.name not in caplog.text


def test_set_samples_by_valid_case_id_and_valid_identifier(
    cli_runner, base_context, base_store: Store, helpers, caplog
):
    # GIVEN a database with a sample that belongs to a case
    case_obj = helpers.add_case(store=base_store)
    sample_obj = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_obj)

    # WHEN calling set samples with case_id for sample and valid identifier for sample
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples,
            [
                case_obj.internal_id,
                "-id",
                "internal_id",
                sample_obj.internal_id,
                "-y",
                "--skip-lims",
            ],
            obj=base_context,
        )

    # THEN it should name the sample to be changed
    assert sample_obj.internal_id in caplog.text
    assert sample_obj.name in caplog.text
    assert result.exit_code == SUCCESS


def test_set_samples_by_invalid_case_id_and_valid_identifier(
    cli_runner, base_context: CGConfig, helpers, caplog
):
    # GIVEN a database with a sample that belongs to a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(store=base_store)
    sample_obj = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_obj)

    # WHEN calling set samples with bad case_id for sample and valid identifier for sample
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples,
            ["wrong_caseid", "-id", "internal_id", sample_obj.internal_id, "-y", "--skip-lims"],
            obj=base_context,
        )

    # THEN it should not name the sample to be changed
    assert sample_obj.internal_id not in caplog.text
    assert sample_obj.name not in caplog.text
    assert result.exit_code != SUCCESS


def test_set_samples_by_valid_case_id_and_invalid_identifier(
    cli_runner, base_context: CGConfig, helpers, caplog
):
    # GIVEN a database with a sample that belongs to a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(store=base_store)
    sample_obj = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_obj)

    # WHEN calling set samples with valid case_id for sample and wrong valid identifier for sample
    with caplog.at_level(logging.INFO):
        result = cli_runner.invoke(
            samples,
            [case_obj.internal_id, "-id", "internal_id", "wrong_internal_id", "-y", "--skip-lims"],
            obj=base_context,
        )

    # THEN it should not name the sample to be changed
    assert sample_obj.internal_id not in caplog.text
    assert sample_obj.name not in caplog.text
    assert result.exit_code != SUCCESS



from cg.services.delivery_message.messages.analysis_scout_message import AnalysisScoutMessage
from cg.services.delivery_message.messages.covid_message import CovidMessage
from cg.services.delivery_message.messages.fastq_message import FastqMessage
from cg.services.delivery_message.messages.fastq_scout_message import FastqScoutMessage
from cg.services.delivery_message.messages.fastq_analysis_scout_message import (
    FastqAnalysisScoutMessage,
)
from cg.services.delivery_message.messages.microsalt_mwr_message import MicrosaltMwrMessage
from cg.services.delivery_message.messages.microsalt_mwx_message import MicrosaltMwxMessage
from cg.services.delivery_message.messages.scout_message import ScoutMessage
from cg.services.delivery_message.messages.statina_message import StatinaMessage



"""Module that holds the illumina post-processing service."""

import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.devices import DeviceType
from cg.exc import MissingFilesError, FlowCellError
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.services.illumina_post_processing_service.utils import (
    create_delivery_file_in_flow_cell_directory,
)
from cg.services.illumina_post_processing_service.validation import (
    is_flow_cell_ready_for_postprocessing,
)
from cg.store.models import IlluminaFlowCell
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class IlluminaPostProcessingService:
    def __init__(self, status_db: Store, housekeeper_api: HousekeeperAPI, dry_run: bool) -> None:
        self.status_db: Store = status_db
        self.hk_api: HousekeeperAPI = housekeeper_api
        self.dry_run: bool = False

    @staticmethod
    def store_illumina_flow_cell(
        flow_cell: FlowCellDirectoryData,
        store: Store,
    ) -> IlluminaFlowCell:
        """
        Create flow cell from the parsed and validated flow cell data.
        And add the samples on the flow cell to the model.
        """
        model: str | None = flow_cell.run_parameters.get_flow_cell_model()
        new_flow_cell = IlluminaFlowCell(
            internal_id=flow_cell.id, type=DeviceType.ILLUMINA, model=model
        )
        return store.add_illumina_flow_cell(new_flow_cell)

    @staticmethod
    def store_illumina_sequencing_metrics(flow_cell: IlluminaFlowCell) -> None:
        """Store illumina run metrics in the status database."""
        pass

    @staticmethod
    def store_illumina_sample_sequencing_metrics():
        """Store illumina sample sequencing metrics in the status database."""
        pass

    def store_illumina_flow_cell_data(self, flow_cell: FlowCellDirectoryData) -> None:
        """Store flow cell data in the status database."""
        flow_cell: IlluminaFlowCell = self.store_illumina_flow_cell(
            flow_cell=flow_cell, store=self.status_db
        )
        self.store_illumina_sequencing_metrics(flow_cell)
        self.store_illumina_sample_sequencing_metrics()
        self.status_db.commit_to_store()

    def post_process_illumina_flow_cell(
        self,
        flow_cell_directory_name: str,
        demultiplexed_runs_dir: Path,
    ) -> None:
        """Store data for the demultiplexed flow cell and mark it as ready for delivery.
        This function:
            - Stores the flow cell in the status database
            - Stores sequencing metrics in the status database
            - Updates sample read counts in the status database
            - Stores the flow cell data in the housekeeper database
            - Creates a delivery file in the flow cell directory
        Raises:
            FlowCellError: If the flow cell directory or the data it contains is not valid.
        """

        LOG.info(f"Post-process flow cell {flow_cell_directory_name}")
        flow_cell_out_directory = Path(demultiplexed_runs_dir, flow_cell_directory_name)
        flow_cell = FlowCellDirectoryData(flow_cell_out_directory)
        sample_sheet_path: Path = self.hk_api.get_sample_sheet_path(flow_cell.id)
        flow_cell.set_sample_sheet_path_hk(hk_path=sample_sheet_path)

        LOG.debug("Set path for Housekeeper sample sheet in flow cell")
        try:
            is_flow_cell_ready_for_postprocessing(
                flow_cell_output_directory=flow_cell_out_directory,
                flow_cell=flow_cell,
            )
        except (FlowCellError, MissingFilesError) as e:
            LOG.warning(f"Flow cell {flow_cell_directory_name} will be skipped: {e}")
            return
        if self.dry_run:
            LOG.info(f"Dry run will not finish flow cell {flow_cell_directory_name}")
            return
        try:
            self.store_illumina_flow_cell_data(flow_cell)
        except Exception as e:
            LOG.error(f"Failed to store flow cell data: {str(e)}")
            raise
        create_delivery_file_in_flow_cell_directory(flow_cell_out_directory)



import mock

from cg.constants.invoice import CostCenters
from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import RecordType
from cg.meta.invoice import InvoiceAPI
from cg.models.invoice.invoice import InvoiceInfo
from cg.store.models import Pool, Sample


def test_instantiate_invoice_api(get_invoice_api_sample):
    # GIVEN a sample invoice api

    # When instantiating
    sample_invoice_api: InvoiceAPI = get_invoice_api_sample

    # THEN return an InvoiceAPI instance
    assert isinstance(sample_invoice_api, InvoiceAPI)


def test_assert_invoice_api(get_invoice_api_sample):
    # GIVEN an invoice API
    api: InvoiceAPI = get_invoice_api_sample
    # THEN a report should be generated
    report = api.get_invoice_report(CostCenters.ki)
    assert type(report) == dict


def test_invoice_api_sample(get_invoice_api_sample, record_type: str = RecordType.Sample):
    """Test that the invoice records the right record_type"""
    # THEN calling InvoiceAPI should return an API
    api: InvoiceAPI = get_invoice_api_sample
    assert api
    # THEN record_type should be Sample
    assert api.record_type == record_type
    # THEN prepare_invoice_report should set priority to standard
    api.get_invoice_report(CostCenters.ki)
    assert api.invoice_info.priority == PriorityTerms.STANDARD

    # THEN get_invoice_report returns a dictionary
    invoice_dict: dict = api.get_invoice_report(CostCenters.ki)
    assert type(invoice_dict) == dict

    # THEN prepare_invoice_report should set priority to standard
    assert api.invoice_info.priority == PriorityTerms.STANDARD

    # THEN raw.records should contain a Pool
    assert type(api.raw_records[0]) == Sample

    # THEN api holds an InvoiceInfo class
    assert type(api.invoice_info) == InvoiceInfo


def test_invoice_api_nipt_customer(
    get_invoice_api_nipt_customer, record_type: str = RecordType.Pool
):
    # GIVEN an invoice api with NIPT customer
    api: InvoiceAPI = get_invoice_api_nipt_customer

    # THEN record_type should be a Pool
    api.genologics_lims = mock.MagicMock()
    assert api.record_type == record_type

    # THEN get_invoice_report returns a dictionary
    invoice_dict = api.get_invoice_report(CostCenters.ki)
    assert type(invoice_dict) == dict

    # THEN prepare_invoice_report should set priority to standard
    assert api.invoice_info.priority == PriorityTerms.STANDARD

    # THEN raw.records should contain a Pool
    assert type(api.raw_records[0]) == Pool

    # THEN api holds an InvoiceInfo class
    assert type(api.invoice_info) == InvoiceInfo


def test_invoice_pool_generic_customer(
    get_invoice_api_pool_generic_customer, record_type: str = RecordType.Pool
):
    # GIVEN an invoice API with a pool and a generic customer
    api: InvoiceAPI = get_invoice_api_pool_generic_customer
    assert api
    # THEN record_type should be Pool
    api.genologics_lims = mock.MagicMock()
    assert api.record_type == record_type
    # THEN prepare_invoice_report should set priority to research
    api.get_invoice_report(CostCenters.ki)
    assert api.invoice_info.priority == PriorityTerms.RESEARCH



import logging

import click
from cg.models.cg_config import CGConfig
from cg.services.validate_file_transfer_service.validate_pacbio_file_transfer_service import (
    ValidatePacbioFileTransferService,
)


LOG = logging.getLogger(__name__)


@click.group()
def validate():
    """Validation of processes in cg."""


@validate.command("pacbio-transfer")
@click.pass_obj
def validate_pacbio_transfer(context: CGConfig):
    """Validate that the PacBio transfer is correct."""
    validate_service = ValidatePacbioFileTransferService(config=context)
    validate_service.validate_all_transfer_done()



"""Tests for BALSAMIC analysis."""

from pathlib import Path

import pytest

from cg.constants.observations import ObservationsFileWildcards
from cg.constants.sequencing import Variants
from cg.constants.subject import Sex
from cg.exc import BalsamicStartError
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.cg_config import CGConfig
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis


def test_get_verified_sex():
    """Test sex extraction from a sample dictionary."""

    # GIVEN a sample object
    sample_obj = {
        "ACC0000A0": {"sex": "female"},
        "ACC0000A1": {"sex": "female"},
    }

    # WHEN extracting the sex
    retrieved_sex: Sex = BalsamicAnalysisAPI.get_verified_sex(sample_obj)

    # THEN sex must match the expected one
    assert retrieved_sex == "female"


def test_get_verified_sex_error():
    """Test sex extraction from a sample dictionary when two different sexes have been provided."""

    # GIVEN a sample object with different sexes
    sample_obj = {
        "ACC0000A0": {"sex": "male"},
        "ACC0000A1": {"sex": "female"},
    }

    # WHEN extracting the sex
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_sex(sample_obj)


def test_get_verified_sex_unknown(caplog):
    """Tests sex extraction from a sample dictionary when the sex is unknown."""

    # GIVEN a sample object with different sexes
    sample_obj = {
        "ACC0000A0": {"sex": "unknown"},
        "ACC0000A1": {"sex": "unknown"},
    }

    # WHEN extracting the sex
    retrieved_sex: Sex = BalsamicAnalysisAPI.get_verified_sex(sample_obj)

    # THEN sex must match the expected one
    assert retrieved_sex == Sex.FEMALE
    assert f"The provided sex is unknown, setting {Sex.FEMALE} as the default" in caplog.text


def test_get_verified_pon():
    """Tests PON verification."""

    # GIVEN specific panel bed and PON files
    panel_bed = "/path/gmcksolid_4.1_hg19_design.bed"
    pon_cnn = "/path/PON/gmcksolid_4.1_hg19_design_CNVkit_PON_reference_v2.cnn"
    invalid_pon_cnn = "/path/PON/gmssolid_15.2_hg19_design_CNVkit_PON_reference_v2.cnn"

    # WHEN validating the PON
    validated_pon: str = BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, pon_cnn)

    # THEN the PON verification should be performed successfully
    assert pon_cnn == validated_pon
    with pytest.raises(BalsamicStartError):
        BalsamicAnalysisAPI.get_verified_pon(None, panel_bed, invalid_pon_cnn)


def test_get_latest_file_by_pattern(
    cg_context: CGConfig, observations_dir: Path, observations_somatic_snv_file_path: Path
):
    """Test latest observations extraction."""

    # GIVEN a Loqusdb temporary directory and a cancer SNV file wildcard
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)

    # WHEN getting the latest observations file
    observation: str = balsamic_analysis_api.get_latest_file_by_pattern(
        directory=observations_dir, pattern=ObservationsFileWildcards.CANCER_SOMATIC_SNV
    )

    # THEN the extracted observation should match the latest file
    assert observation == observations_somatic_snv_file_path.as_posix()


def test_get_parsed_observation_file_paths_no_args(
    cg_context: CGConfig,
    observations_dir: Path,
    observations_clinical_sv_file_path: Path,
    observations_somatic_snv_file_path: Path,
    outdated_observations_somatic_snv_file_path: Path,
):
    """Test verified observations extraction with no arguments."""

    # GIVEN a Loqusdb temporary directory
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.loqusdb_path = observations_dir

    # WHEN getting the latest observations arguments dictionary
    args: dict = balsamic_analysis_api.get_parsed_observation_file_paths(None)

    # THEN only the created observations files should be returned
    assert args[ObservationsFileWildcards.CLINICAL_SNV] is None
    assert (
        args[ObservationsFileWildcards.CLINICAL_SV] == observations_clinical_sv_file_path.as_posix()
    )
    assert args[ObservationsFileWildcards.CANCER_GERMLINE_SNV] is None
    assert args[ObservationsFileWildcards.CANCER_GERMLINE_SV] is None
    assert (
        args[ObservationsFileWildcards.CANCER_SOMATIC_SNV]
        == observations_somatic_snv_file_path.as_posix()
    )
    assert (
        outdated_observations_somatic_snv_file_path.as_posix()
        not in args[ObservationsFileWildcards.CANCER_SOMATIC_SNV]
    )
    assert args[ObservationsFileWildcards.CANCER_SOMATIC_SV] is None


def test_get_parsed_observation_file_paths_overwrite_input(
    cg_context: CGConfig,
    observations_dir: Path,
    observations_clinical_snv_file_path: Path,
    custom_observations_clinical_snv_file_path: Path,
):
    """Test verified observations extraction when providing a non default observation file."""

    # GIVEN a Loqusdb temporary directory and a custom Loqusdb file
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.loqusdb_path = observations_dir

    # WHEN getting the latest observations dictionary
    args: dict = balsamic_analysis_api.get_parsed_observation_file_paths(
        [custom_observations_clinical_snv_file_path.as_posix()]
    )

    # THEN the default file should be overwritten by the custom file
    assert (
        observations_clinical_snv_file_path.as_posix()
        not in args[ObservationsFileWildcards.CLINICAL_SNV]
    )
    assert (
        args[ObservationsFileWildcards.CLINICAL_SNV]
        == custom_observations_clinical_snv_file_path.as_posix()
    )


def test_get_swegen_verified_path(
    cg_context: CGConfig, swegen_dir: Path, swegen_snv_reference: Path
):
    """Test verified SweGen path return."""

    # GIVEN A SNV variants pattern and a BALSAMIC analysis API
    balsamic_analysis_api: BalsamicAnalysisAPI = BalsamicAnalysisAPI(cg_context)
    balsamic_analysis_api.swegen_path = swegen_dir

    # WHEN obtaining the latest file by pattern
    swegen_file: str = balsamic_analysis_api.get_swegen_verified_path(Variants.SNV)

    # THEN the returned file should match the SweGen SNV reference
    assert swegen_file == swegen_snv_reference.as_posix()


def test_get_variant_callers(analysis_api_balsamic: MockBalsamicAnalysis, case_id: str):
    """Test variant callers extraction for TN-PANEL analysis."""

    # GIVEN a mock metadata object
    balsamic_metadata: BalsamicAnalysis = analysis_api_balsamic.get_latest_metadata(case_id)

    # GIVEN an expected output for tumor normal panel analysis
    expected_callers = [
        "manta (v1.6.0)",
        "cnvkit",
        "vardict (v2019.06.04=pl526_0)",
        "dnascope",
        "tnhaplotyper",
        "manta_germline (v1.6.0)",
        "haplotypecaller",
        "TNscope_umi",
        "delly (v0.8.7)",
    ]

    # WHEN retrieving the analysis variant callers
    variant_callers: list[str] = analysis_api_balsamic.get_variant_callers(balsamic_metadata)

    # THEN check that the callers are correctly identified
    assert variant_callers == expected_callers


def test_get_variant_caller_version(analysis_api_balsamic: MockBalsamicAnalysis, case_id: str):
    """Tests variant caller version extraction."""

    # GIVEN a tool name and a mock variant caller versions dictionary
    var_caller_name = "manta"
    var_caller_versions: dict[str, list[str]] = analysis_api_balsamic.get_latest_metadata(
        case_id
    ).config.bioinfo_tools_version

    # GIVEN the tools mock version
    expected_version = "1.6.0"

    # WHEN retrieving the version of a specific variant caller
    version: str | None = analysis_api_balsamic.get_variant_caller_version(
        var_caller_name, var_caller_versions
    )

    # THEN verify that the extracted version is correct
    assert version == expected_version



from pydantic.v1 import BaseModel

from cg.models.rnafusion.rnafusion import RnafusionQCMetrics
from cg.models.taxprofiler.taxprofiler import TaxprofilerQCMetrics
from cg.models.tomte.tomte import TomteQCMetrics


class AnalysisModel(BaseModel):
    """Metadata analysis model"""


class NextflowAnalysis(AnalysisModel):
    """Nextflow's analysis results model."""

    sample_metrics: dict[str, RnafusionQCMetrics | TaxprofilerQCMetrics | TomteQCMetrics]



"""Tests utils for generating delivery reports"""

from datetime import datetime

import click.exceptions
import pytest

from cg.cli.generate.report.utils import (
    get_report_analysis_started,
    get_report_api,
    get_report_api_workflow,
    get_report_case,
)
from cg.constants import Workflow
from cg.meta.report.balsamic_umi import BalsamicUmiReportAPI
from tests.mocks.report import MockMipDNAReportAPI


def test_get_report_case(delivery_report_click_context, case_id):
    """Tests case object extraction for delivery report given a valid case_id"""

    # GIVEN a delivery report click context

    # WHEN resolving a report case ID
    case_obj = get_report_case(delivery_report_click_context, case_id)

    # THEN a valid case object should be returned
    assert case_obj.internal_id == case_id


def test_get_report_case(delivery_report_click_context, caplog):
    """Tests error raise when an incorrect case ID is provided"""

    # GIVEN a delivery report click context

    # WHEN resolving an incorrect case ID, an exception should be raised
    with pytest.raises(click.exceptions.Abort):
        get_report_case(delivery_report_click_context, "not a case ID")

    assert "Invalid case ID" in caplog.text


def test_get_report_api(delivery_report_click_context, cg_context, case_id):
    """Tests report API extraction"""

    # GIVEN a case object
    case_obj = get_report_case(delivery_report_click_context, case_id)

    # WHEN retrieving a report API
    report_api = get_report_api(delivery_report_click_context, case_obj)

    # THEN the extracted API should be correctly retrieved
    assert report_api
    assert isinstance(report_api, MockMipDNAReportAPI)


def test_get_report_api_workflow(delivery_report_click_context):
    """Tests API assignment given a specific workflow."""

    # GIVEN a click context and a specific workflow
    workflow = Workflow.BALSAMIC_UMI

    # WHEN validating a report api
    report_api = get_report_api_workflow(delivery_report_click_context, workflow)

    # THEN the extracted API should match the expected type
    assert report_api
    assert isinstance(report_api, BalsamicUmiReportAPI)


def test_get_report_analysis_started(delivery_report_click_context, case_id):
    """Tests retrieval of analysis started at field"""

    # GIVEN a case object and a report api
    case_obj = get_report_case(delivery_report_click_context, case_id)
    report_api = get_report_api(delivery_report_click_context, case_obj)

    # WHEN resolving the analysis started at field
    started_at = get_report_analysis_started(case_obj, report_api, None)

    # THEN check if the verified value has been correctly extracted and formatted
    assert started_at
    assert isinstance(started_at, datetime)



from enum import StrEnum

from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.models.orders.order import OrderType

SEX_MAP = {"male": "M", "female": "F", "unknown": "unknown"}
REV_SEX_MAP = {value: key for key, value in SEX_MAP.items()}
CONTAINER_TYPES = ["Tube", "96 well plate"]
SOURCE_TYPES = set().union(METAGENOME_SOURCES, ANALYSIS_SOURCES)

CASE_PROJECT_TYPES = [
    OrderType.MIP_DNA,
    OrderType.BALSAMIC,
    OrderType.MIP_RNA,
]


class Orderform(StrEnum):
    BALSAMIC: str = "1508"
    BALSAMIC_QC: str = "1508"
    BALSAMIC_UMI: str = "1508"
    FASTQ: str = "1508"
    METAGENOME: str = "1508"
    FLUFFY: str = "1604"
    MICROSALT: str = "1603"
    MIP_DNA: str = "1508"
    MIP_RNA: str = "1508"
    RNAFUSION: str = "1508"
    RML: str = "1604"
    SARS_COV_2: str = "2184"

    @staticmethod
    def get_current_orderform_version(order_form: str) -> str:
        """Returns the current version of the given order form."""
        current_order_form_versions = {
            Orderform.MIP_DNA: "31",
            Orderform.RML: "17",
            Orderform.MICROSALT: "11",
            Orderform.SARS_COV_2: "9",
        }
        return current_order_form_versions[order_form]



"""Test delivery API methods."""

import logging
from pathlib import Path

from _pytest.logging import LogCaptureFixture
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import DataDelivery
from cg.constants.delivery import INBOX_NAME
from cg.meta.delivery.delivery import DeliveryAPI
from cg.models.cg_config import CGConfig
from cg.models.delivery.delivery import DeliveryFile
from cg.store.models import Case, Sample
from cg.store.store import Store


def test_is_sample_deliverable(delivery_context_microsalt: CGConfig, sample_id: str):
    """Test that a sample is deliverable."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN a sample id with enough reads
    sample: Sample = status_db.get_sample_by_internal_id(sample_id)

    # WHEN evaluating if sample is deliverable
    is_deliverable: bool = delivery_api.is_sample_deliverable(sample)

    # THEN the sample should be deliverable
    assert is_deliverable


def test_is_sample_deliverable_false(
    delivery_context_microsalt: CGConfig, sample_id_not_enough_reads: str
):
    """Test that a sample is not deliverable."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN a sample without enough reads
    sample: Sample = status_db.get_sample_by_internal_id(sample_id_not_enough_reads)

    # WHEN evaluating if sample is deliverable
    is_deliverable: bool = delivery_api.is_sample_deliverable(sample)

    # THEN the sample should not be deliverable
    assert not is_deliverable


def test_is_sample_deliverable_force(
    delivery_context_microsalt: CGConfig, sample_id_not_enough_reads: str
):
    """Test that a non-deliverable sample is deliverable with a force flag."""
    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN a rare disease sample without enough reads
    sample: Sample = status_db.get_sample_by_internal_id(sample_id_not_enough_reads)

    # WHEN evaluating if sample is deliverable with a force flag
    is_deliverable: bool = delivery_api.is_sample_deliverable(sample=sample, force=True)

    # THEN the sample should be deliverable
    assert is_deliverable


def test_convert_case_files_to_delivery_files(delivery_context_balsamic: CGConfig, case_id: str):
    """Test Housekeeper case files conversion to delivery files."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    housekeeper_api: HousekeeperAPI = delivery_context_balsamic.housekeeper_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case object and list of Housekeeper files to be delivered
    case: Case = status_db.get_case_by_internal_id(case_id)
    hk_files: list[File] = housekeeper_api.get_files(bundle=case_id, tags=[case_id]).all()

    # WHEN converting Housekeeper case files to delivery files
    delivery_files: list[DeliveryFile] = delivery_api.convert_files_to_delivery_files(
        files=hk_files, case=case, internal_id=case.internal_id, external_id=case.name
    )

    # THEN the delivery files should be correctly formatted
    hk_file: File = hk_files[0]
    delivery_file: DeliveryFile = delivery_files[0]
    destination_dir = Path(
        delivery_api.delivery_path,
        case.customer.internal_id,
        INBOX_NAME,
        case.latest_ticket,
        case.name,
    )
    assert delivery_file.source_path.as_posix() == hk_file.full_path
    assert delivery_file.destination_path.parent == destination_dir
    assert case.name in delivery_file.destination_path.as_posix()
    assert case.internal_id not in delivery_file.destination_path.as_posix()


def test_convert_sample_files_to_delivery_files(
    delivery_context_balsamic: CGConfig, case_id: str, sample_id: str
):
    """Test Housekeeper sample files conversion to delivery files."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    housekeeper_api: HousekeeperAPI = delivery_context_balsamic.housekeeper_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case, a sample object, and a list of sample Housekeeper files to be delivered
    case: Case = status_db.get_case_by_internal_id(case_id)
    sample: Sample = status_db.get_sample_by_internal_id(sample_id)
    hk_sample_files: list[File] = housekeeper_api.get_files(bundle=case_id, tags=[sample_id]).all()

    # WHEN converting Housekeeper sample files to delivery files
    delivery_files: list[DeliveryFile] = delivery_api.convert_files_to_delivery_files(
        files=hk_sample_files,
        case=case,
        internal_id=sample.internal_id,
        external_id=sample.name,
        analysis_sample_files=True,
    )

    # THEN the delivery sample files should be correctly formatted
    hk_file: File = hk_sample_files[0]
    delivery_file: DeliveryFile = delivery_files[0]
    destination_dir = Path(
        delivery_api.delivery_path,
        case.customer.internal_id,
        INBOX_NAME,
        case.latest_ticket,
        case.name,
        sample.name,
    )
    assert delivery_file.source_path.as_posix() == hk_file.full_path
    assert delivery_file.destination_path.parent == destination_dir
    assert sample.name in delivery_file.destination_path.as_posix()
    assert sample.internal_id not in delivery_file.destination_path.as_posix()


def test_convert_sample_files_to_delivery_files_with_case_id(
    delivery_context_balsamic: CGConfig, case_id: str, sample_id: str
):
    """
    Test Housekeeper sample files conversion to delivery files when sample files have case ID in
    their name.
    """

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    housekeeper_api: HousekeeperAPI = delivery_context_balsamic.housekeeper_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case, a sample object, and a list of sample Housekeeper files with case ID in their
    # name to be delivered
    case: Case = status_db.get_case_by_internal_id(case_id)
    sample: Sample = status_db.get_sample_by_internal_id(sample_id)
    hk_sample_files: list[File] = housekeeper_api.get_files(bundle=case_id, tags=[sample_id]).all()
    hk_case_files: list[File] = housekeeper_api.get_files(bundle=case_id, tags=[case_id]).all()
    hk_files: list[File] = hk_sample_files + hk_case_files

    # WHEN converting Housekeeper sample files to delivery files
    delivery_files: list[DeliveryFile] = delivery_api.convert_files_to_delivery_files(
        files=hk_files,
        case=case,
        internal_id=sample.internal_id,
        external_id=sample.name,
        analysis_sample_files=True,
    )

    print(delivery_files)

    # THEN the delivery sample files should be correctly formatted and not contain case or sample
    # internal IDs
    for delivery_file in delivery_files:
        assert sample.name in delivery_file.destination_path.as_posix()
        assert sample.internal_id not in delivery_file.destination_path.as_posix()
        assert case.internal_id not in delivery_file.destination_path.as_posix()


def test_get_fastq_delivery_files_by_sample(
    delivery_context_microsalt: CGConfig,
    case_id: str,
    sample_id: str,
    delivery_fastq_file: Path,
    delivery_spring_file: Path,
):
    """Test get FASTQ delivery files for a sample."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN case and sample objects
    case: Case = status_db.get_case_by_internal_id(case_id)
    sample: Sample = status_db.get_sample_by_internal_id(sample_id)

    # WHEN retrieving FASTQ delivery files by sample
    delivery_files: list[DeliveryFile] = delivery_api.get_fastq_delivery_files_by_sample(
        case=case, sample=sample
    )

    # THEN only FASTQ files should be returned
    assert isinstance(delivery_files[0], DeliveryFile)
    assert delivery_files[0].source_path.name == delivery_fastq_file.name
    for delivery_file in delivery_files:
        assert delivery_file.source_path.name != delivery_spring_file.name


def test_get_fastq_delivery_files_by_sample_not_deliverable(
    delivery_context_microsalt: CGConfig,
    case_id: str,
    sample_id_not_enough_reads: str,
    caplog: LogCaptureFixture,
):
    """Test get FASTQ delivery files for a sample that is not deliverable."""
    caplog.set_level(logging.INFO)

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN case and sample objects
    case: Case = status_db.get_case_by_internal_id(case_id)
    sample: Sample = status_db.get_sample_by_internal_id(sample_id_not_enough_reads)

    # WHEN retrieving FASTQ delivery files by a sample that is not deliverable
    delivery_files: list[DeliveryFile] = delivery_api.get_fastq_delivery_files_by_sample(
        case=case, sample=sample
    )

    # THEN no delivery files should be returned
    assert not delivery_files
    assert f"Sample {sample_id_not_enough_reads} is not deliverable" in caplog.text


def test_get_fastq_delivery_files(
    delivery_context_microsalt: CGConfig,
    case_id: str,
    delivery_fastq_file: Path,
    delivery_another_fastq_file: Path,
):
    """Test get FASTQ delivery files for all samples linked to a case."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN a case object
    case: Case = status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving the FASTQ file to deliver
    delivery_files: list[DeliveryFile] = delivery_api.get_fastq_delivery_files(case=case)

    # THEN all the FASTQ sample files should be returned
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name in [
            delivery_fastq_file.name,
            delivery_another_fastq_file.name,
        ]


def test_get_analysis_sample_delivery_files_by_sample(
    delivery_context_balsamic: CGConfig, case_id: str, sample_id: str, delivery_cram_file: Path
):
    """Test get analysis files to deliver by sample."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN case and sample analysis objects
    case: Case = status_db.get_case_by_internal_id(case_id)
    sample: Sample = status_db.get_sample_by_internal_id(sample_id)

    # WHEN retrieving delivery files by sample
    delivery_files: list[DeliveryFile] = delivery_api.get_analysis_sample_delivery_files_by_sample(
        case=case, sample=sample
    )

    # THEN the analysis cram file should be returned as a delivery file model
    assert isinstance(delivery_files[0], DeliveryFile)
    assert delivery_files[0].source_path.name == delivery_cram_file.name


def test_get_analysis_sample_delivery_files(
    delivery_context_balsamic: CGConfig,
    case_id: str,
    delivery_cram_file: Path,
    delivery_another_cram_file: Path,
):
    """Test get complete list of analysis sample files."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case object
    case: Case = status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving sample delivery files
    delivery_files: list[DeliveryFile] = delivery_api.get_analysis_sample_delivery_files(case=case)

    # THEN the analysis cram files should be returned for all case samples
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name in [
            delivery_cram_file.name,
            delivery_another_cram_file.name,
        ]


def test_get_analysis_case_delivery_files(
    delivery_context_balsamic: CGConfig,
    case_id: str,
    delivery_report_file: Path,
    delivery_cram_file: Path,
    delivery_another_cram_file: Path,
):
    """Test analysis case delivery files retrieval."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case object
    case: Case = status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving case delivery files
    delivery_files: list[DeliveryFile] = delivery_api.get_analysis_case_delivery_files(case=case)

    # THEN only case specific files should be returned, ignoring sample analysis files
    assert delivery_files[0].source_path.name == delivery_report_file.name
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name not in [
            delivery_cram_file.name,
            delivery_another_cram_file.name,
        ]


def test_get_delivery_files_fastq_delivery(
    delivery_context_microsalt: CGConfig,
    case_id: str,
    delivery_fastq_file: Path,
    delivery_another_fastq_file: Path,
):
    """Test get delivery files for FASTQ data delivery."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_microsalt.delivery_api
    status_db: Store = delivery_context_microsalt.status_db

    # GIVEN a case object with FASTQ data as delivery
    case: Case = status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving the FASTQ delivery files
    delivery_files: list[DeliveryFile] = delivery_api.get_delivery_files(case=case)

    # THEN only the FASTQ sample files should be returned
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name in [
            delivery_fastq_file.name,
            delivery_another_fastq_file.name,
        ]


def test_get_delivery_files_analysis_delivery(
    delivery_context_balsamic: CGConfig,
    case_id: str,
    delivery_report_file: Path,
    delivery_cram_file: Path,
    delivery_another_cram_file: Path,
):
    """Test get delivery files for analysis data delivery."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case object with analysis files as data delivery
    case: Case = status_db.get_case_by_internal_id(case_id)
    case.data_delivery = DataDelivery.ANALYSIS_FILES

    # WHEN retrieving the analysis delivery files
    delivery_files: list[DeliveryFile] = delivery_api.get_delivery_files(case=case)

    # THEN only the analysis case and sample files should be returned
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name in [
            delivery_report_file.name,
            delivery_cram_file.name,
            delivery_another_cram_file.name,
        ]


def test_get_delivery_files_fastq_analysis_delivery(
    delivery_context_balsamic: CGConfig,
    case_id: str,
    delivery_report_file: Path,
    delivery_cram_file: Path,
    delivery_another_cram_file: Path,
    delivery_fastq_file: Path,
    delivery_another_fastq_file: Path,
):
    """Test get delivery files for FASTQ analysis data delivery."""

    # GIVEN a delivery context
    delivery_api: DeliveryAPI = delivery_context_balsamic.delivery_api
    status_db: Store = delivery_context_balsamic.status_db

    # GIVEN a case object with FASTQ analysis as data delivery
    case: Case = status_db.get_case_by_internal_id(case_id)

    # WHEN retrieving the FASTQ analysis delivery files
    delivery_files: list[DeliveryFile] = delivery_api.get_delivery_files(case=case)

    # THEN analysis case and sample files should be returned together with the fastqs
    for delivery_file in delivery_files:
        assert isinstance(delivery_file, DeliveryFile)
        assert delivery_file.source_path.name in [
            delivery_report_file.name,
            delivery_cram_file.name,
            delivery_another_cram_file.name,
            delivery_fastq_file.name,
            delivery_another_fastq_file.name,
        ]



"""Utility functions for the Illumina post-processing service."""

from pathlib import Path
from xml.etree.ElementTree import ElementTree, Element

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles, RunParametersXMLNodes
from cg.exc import XMLError
from cg.io.xml import read_xml, get_tree_node


def create_delivery_file_in_flow_cell_directory(flow_cell_directory: Path) -> None:
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).touch()



from typing import Callable

from cg.store.models import Case, Sample
from cg.services.quality_controller.quality_checks.checks import (
    run_quality_checks,
    get_sequencing_quality_check_for_case,
    get_sample_sequencing_quality_check,
)


class QualityControllerService:
    """Quality controller class. This class is used to run the quality checks for the samples and cases."""

    @staticmethod
    def case_pass_sequencing_qc(case: Case) -> bool:
        """
        Run the QC for the case or sample.
        """
        sequencing_quality_check: Callable = get_sequencing_quality_check_for_case(case)
        return run_quality_checks(quality_checks=[sequencing_quality_check], case=case)

    @staticmethod
    def sample_pass_sequencing_qc(sample: Sample) -> bool:
        """
        Run the sequencing QC for a sample.
        """
        sample_sequencing_quality_check: Callable = get_sample_sequencing_quality_check()
        return run_quality_checks(quality_checks=[sample_sequencing_quality_check], sample=sample)



"""Loqusdb API fixtures."""

import pytest
from pytest_mock import MockFixture

from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import LOQUSDB_ID
from cg.constants.sample_sources import SourceType
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Customer
from cg.store.store import Store


@pytest.fixture
def balsamic_observations_api(
    cg_context: CGConfig,
    analysis_store: Store,
    lims_api: LimsAPI,
    loqusdb_api: LoqusdbAPI,
    case_id: str,
    balsamic_customer: Customer,
    number_of_loaded_variants: int,
    loqusdb_id: str,
    mocker: MockFixture,
) -> BalsamicObservationsAPI:
    """Cancer observations API fixture."""
    balsamic_observations_api: BalsamicObservationsAPI = BalsamicObservationsAPI(cg_context)
    balsamic_observations_api.store = analysis_store
    balsamic_observations_api.loqusdb_somatic_api = loqusdb_api
    balsamic_observations_api.loqusdb_tumor_api = loqusdb_api

    # Mocked case scenario for Balsamic uploads
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    case.customer.internal_id = balsamic_customer.internal_id
    case.samples[0].is_tumour = True

    # Mocked Loqusdb API scenario for Balsamic uploads
    mocker.patch.object(LoqusdbAPI, "load", return_value={"variants": number_of_loaded_variants})
    mocker.patch.object(
        LoqusdbAPI, "get_case", return_value={"case_id": case_id, LOQUSDB_ID: loqusdb_id}
    )

    # Mocked LIMS API scenario
    mocker.patch.object(LimsAPI, "get_source", return_value=SourceType.TISSUE)

    return balsamic_observations_api


@pytest.fixture
def mip_dna_observations_api(
    cg_context: CGConfig,
    analysis_store: Store,
    lims_api: LimsAPI,
    loqusdb_api: LoqusdbAPI,
    case_id: str,
    mip_dna_customer: Customer,
    number_of_loaded_variants: int,
    loqusdb_id: str,
    mocker: MockFixture,
) -> MipDNAObservationsAPI:
    """Rare diseases observations API fixture."""
    mip_dna_observations_api: MipDNAObservationsAPI = MipDNAObservationsAPI(cg_context)
    mip_dna_observations_api.store = analysis_store
    mip_dna_observations_api.loqusdb_api = loqusdb_api

    # Mocked case scenario for MIP-DNA uploads
    case: Case = analysis_store.get_case_by_internal_id(case_id)
    case.customer.internal_id = mip_dna_customer.internal_id

    # Mocked Loqusdb API scenario for MIP-DNA uploads
    mocker.patch.object(LoqusdbAPI, "load", return_value={"variants": number_of_loaded_variants})
    mocker.patch.object(
        LoqusdbAPI, "get_case", return_value={"case_id": case_id, LOQUSDB_ID: loqusdb_id}
    )

    # Mocked LIMS API scenario
    mocker.patch.object(LimsAPI, "get_source", return_value=SourceType.TISSUE)

    return mip_dna_observations_api



import pytest
from cg.store.models import Case, Sample
from cg.constants.constants import PrepCategory
from cg.services.quality_controller.quality_checks.utils import (
    any_sample_in_case_has_reads,
    is_case_express_priority,
    case_pass_sequencing_qc,
    express_case_pass_sequencing_qc,
    express_sample_has_enough_reads,
    get_express_reads_threshold_for_sample,
    is_sample_ready_made_library,
    is_sample_express_priority,
    ready_made_library_sample_has_enough_reads,
    sample_has_enough_reads,
)
from cg.store.store import Store
from tests.conftest import StoreHelpers
from tests.fixture_plugins.quality_controller_fixtures.sequencing_qc_check_scenario import (
    SequencingQCCheckScenarios,
)


@pytest.mark.parametrize(
    "sample_fixture, expected_result",
    [
        ("ready_made_library_sample_passing_sequencing_qc", True),
        ("sample_passing_sequencing_qc", False),
    ],
    ids=["ready_made_library", "whole_genome_sequencing"],
)
def test_is_sample_ready_made_library(
    sample_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the is_sample_ready_made_library function.

    This test verifies if the sample is ready-made library based on its prep category.
    """
    # GIVEN a sample with a prep category
    sample: Sample = request.getfixturevalue(sample_fixture)
    # WHEN checking if the sample is a ready-made library
    # THEN the result should be as expected

    assert is_sample_ready_made_library(sample) == expected_result


@pytest.mark.parametrize(
    "case_fixture, expected_result",
    [
        ("express_case_passing_sequencing_qc", True),
        ("case_passing_sequencing_qc", False),
    ],
    ids=["express_priority", "standard_priority"],
)
def test_is_case_express_priority(
    case_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the is_case_express_priority function.

    This test verifies if the case has express priority.
    """
    # GIVEN a case with a priority
    case: Case = request.getfixturevalue(case_fixture)

    # WHEN checking if the case has express priority
    # THEN the result should be as expected
    assert is_case_express_priority(case) == expected_result


@pytest.mark.parametrize(
    "sample_fixture, expected_result",
    [
        ("express_sample_passing_sequencing_qc", True),
        ("sample_passing_sequencing_qc", False),
    ],
    ids=["express_priority", "standard_priority"],
)
def test_is_sample_express_priority(
    sample_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the is_sample_express_priority function.

    This test verifies if the sample has express priority.
    """
    # GIVEN a sample with a priority
    sample: Sample = request.getfixturevalue(sample_fixture)

    # WHEN checking if the sample has express priority
    # THEN the result should be as expected

    assert is_sample_express_priority(sample) == expected_result


@pytest.mark.parametrize(
    "sample_fixture, expected_result",
    [
        ("ready_made_library_sample_passing_sequencing_qc", True),
        ("ready_made_library_sample_failing_sequencing_qc", False),
    ],
    ids=["pass_sequencing_qc", "no_reads"],
)
def test_ready_made_library_sample_has_enough_reads(
    sample_fixture: str,
    expected_result: bool,
    request: pytest.FixtureRequest,
):
    """
    Test the ready_made_library_sample_has_enough_reads function.

    This test verifies if the ready-made library sample has enough reads. It checks if the sample is a ready made
    library and if it has enough reads if so is the case.
    """
    # GIVEN a ready-made library sample with or without reads
    sample: Sample = request.getfixturevalue(sample_fixture)

    # WHEN checking if the ready-made library sample has enough reads
    # THEN the result should be as expected
    assert ready_made_library_sample_has_enough_reads(sample) == expected_result


@pytest.mark.parametrize(
    "sample_fixture, expected_result",
    [
        ("sample_passing_sequencing_qc", True),
        ("sample_failing_sequencing_qc", False),
    ],
    ids=["pass_sequencing_qc", "no_reads"],
)
def test_sample_has_enough_reads(
    sample_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the sample_has_enough_reads function.

    This test verifies if the sample has enough reads. It checks if the sample has enough reads or not.
    """
    # GIVEN a sample with or without reads
    sample: Sample = request.getfixturevalue(sample_fixture)

    # WHEN checking if the sample has enough reads
    # THEN the result should be as expected
    assert sample_has_enough_reads(sample) == expected_result


@pytest.mark.parametrize(
    "case_fixture, expected_result",
    [
        ("case_passing_sequencing_qc", True),
        ("case_failing_sequencing_qc", False),
    ],
    ids=["standard_priority", "standard_priority_no_reads"],
)
def test_get_sequencing_qc_of_case(
    case_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the get_sequencing_qc_of_case function.

    This test verifies the sequencing quality check of the case. It checks if the case has enough
    reads and if it has lower priority than express.

    """
    # GIVEN a case with standard priority and a sample
    case: Case = request.getfixturevalue(case_fixture)
    # WHEN getting the sequencing quality check of the case
    # THEN the sequencing quality check of the case should be as expected

    assert case_pass_sequencing_qc(case) == expected_result


@pytest.mark.parametrize(
    "sample_fixture, expected_result",
    [
        ("express_sample_passing_sequencing_qc", True),
        ("express_sample_failing_sequencing_qc", False),
    ],
    ids=["express_priority", "express_priority_no_reads"],
)
def test_express_sample_has_enough_reads(
    sample_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the express_sample_has_enough_reads function.

    This test verifies the sequencing quality check of the express sample. It checks if the express sample has enough
    reads.

    """
    # GIVEN an express sample with or without reads
    sample: Sample = request.getfixturevalue(sample_fixture)
    # WHEN checking if the express sample has enough reads
    # THEN the result should be as expected

    assert express_sample_has_enough_reads(sample) == expected_result


def test_get_express_reads_threshold_for_sample(express_sample_passing_sequencing_qc: Sample):
    """
    Test the get_express_reads_threshold_for_sample function.

    This test verifies the express reads threshold is correctly calculated for a sample.
    """
    # GIVEN a sample with a target reads value of 10 and an expected threshold of 5
    express_sample_passing_sequencing_qc.application_version.application.target_reads = 10
    expected_express_reads_threshold = 5
    # WHEN getting the express reads threshold for the sample
    express_reads_threshold: int = get_express_reads_threshold_for_sample(
        express_sample_passing_sequencing_qc
    )
    # THEN the express reads threshold should be half of the target reads
    assert express_reads_threshold == expected_express_reads_threshold


@pytest.mark.parametrize(
    "case_fixture, expected_result",
    [
        ("express_case_passing_sequencing_qc", True),
        ("express_case_failing_sequencing_qc", False),
    ],
    ids=["express_priority", "express_priority_no_reads"],
)
def test_express_case_pass_sequencing_qc(
    case_fixture: str, expected_result: bool, request: pytest.FixtureRequest
):
    """
    Test the get_express_sequencing_qc_of_case function.

    This test verifies the express sequencing quality check of the case. It checks if the case has enough
    reads and if it has express priority.
    """

    # GIVEN a case with express priority and a sample
    case: Case = request.getfixturevalue(case_fixture)
    # WHEN getting the express sequencing quality check of the case
    express_sequencing_qc_of_case: bool = express_case_pass_sequencing_qc(case)
    # THEN the express sequencing quality check of the case should be as expected
    assert express_sequencing_qc_of_case == expected_result


@pytest.mark.parametrize(
    "case_fixture, expected_result",
    [
        ("one_sample_in_case_has_reads", True),
        ("no_sample_in_case_has_reads", False),
    ],
    ids=["one_sample_with_reads", "no_sample_with_reads"],
)
def test_any_sample_in_case_has_reads(
    case_fixture: str,
    expected_result: bool,
    request: pytest.FixtureRequest,
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
    base_store: Store,
    helpers: StoreHelpers,
):
    """
    Test the any_sample_in_case_has_reads function.

    This test verifies if any sample in the case has reads.
    """
    # GIVEN a case with a sample with or without reads
    case: Case = request.getfixturevalue(case_fixture)
    # GIVEN that another sample on the case has no reads
    another_sample: Sample = sequencing_qc_check_scenarios.add_sample(
        prep_category=PrepCategory.WHOLE_EXOME_SEQUENCING,
        pass_sequencing_qc=False,
        priority=case.priority,
    )
    helpers.add_relationship(store=base_store, sample=another_sample, case=case)
    # WHEN checking if any sample in the case has reads
    any_sample_in_case_has_reads_result: bool = any_sample_in_case_has_reads(case)
    # THEN the result should be as expected
    assert any_sample_in_case_has_reads_result == expected_result



from datetime import datetime
from pathlib import Path

import pytest

from cg.utils.files import (
    get_directories_in_path,
    get_directory_creation_time_stamp,
    get_file_in_directory,
    remove_directory_and_contents,
    rename_file,
    get_files_in_directory_with_pattern,
)


def test_get_file_in_directory(nested_directory_with_file: Path, some_file: str):
    """Test function to get a file in a directory and subdirectories."""
    # GIVEN a directory with subdirectories with a file

    # WHEN getting the file
    file_path: Path = get_file_in_directory(nested_directory_with_file, some_file)

    # THEN assert that the file is returned
    assert file_path.exists()


def test_get_files_in_directory_by_pattern(nested_directory_with_file: Path, some_file: str):
    """Test function to get files with a pattern in a directory and subdirectories."""
    # GIVEN a directory with subdirectories with a file

    # WHEN getting the file
    file_paths: list[Path] = get_files_in_directory_with_pattern(
        directory=nested_directory_with_file, pattern=some_file
    )

    # THEN assert that the file is returned
    for file_path in file_paths:
        assert file_path.exists()
        assert some_file in file_path.as_posix()


def test_rename_file(tmp_path: Path):
    # GIVEN a file path and a renamed file path
    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path exist
    file_path.touch()
    assert file_path.exists()

    # GIVEN that the renamed file path does not exist
    assert not renamed_file_path.exists()

    # WHEN renaming the file
    rename_file(file_path=file_path, renamed_file_path=renamed_file_path)

    # THEN the renamed file path should exist
    assert renamed_file_path.exists()

    # THEN the file path should not exist
    assert not file_path.exists()


def test_rename_file_exists(tmp_path: Path):
    # GIVEN a file path and a renamed file path
    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path exist
    file_path.touch()
    renamed_file_path.touch()
    assert renamed_file_path.exists()

    # WHEN renaming the file
    rename_file(file_path=file_path, renamed_file_path=renamed_file_path)

    # THEN the renamed file path should exist
    assert renamed_file_path.exists()


def test_rename_file_original_does_not_exist(tmp_path: Path):
    # GIVEN a file path and a renamed file path

    file_path: Path = Path(tmp_path, "dummy_path")
    renamed_file_path: Path = Path(tmp_path, "dummy_renamed_path")

    # GIVEN that the file path does not exist
    assert not file_path.exists()

    # WHEN renaming the file
    # THEN a FileNotFoundError should be raised
    with pytest.raises(FileNotFoundError):
        rename_file(file_path=file_path, renamed_file_path=renamed_file_path)


def test_get_creation_date(tmp_path_factory, timestamp_now: datetime):
    """Test to retrieve the creation date of a directory."""

    # GIVEN a directory that is created now
    directory_path: Path = tmp_path_factory.mktemp("some_dir")

    # WHEN retrieving the creation time stamp
    creation_time_stamp: float = get_directory_creation_time_stamp(directory_path)

    # THEN the creation time stamp is returned
    assert isinstance(creation_time_stamp, float)


def test_get_all_directories_in_path(
    path_with_directories_and_a_file: Path, sub_dir_names: list[str], some_file: str
):
    """Test that get all directories in path only returns directories."""

    # GIVEN a path that contains directories and a file
    assert Path(path_with_directories_and_a_file, some_file).exists()

    # WHEN retrieving all directories in the path
    directories: list[Path] = get_directories_in_path(path=path_with_directories_and_a_file)

    # THEN all directories are returned
    for sub_dir_name in sub_dir_names:
        assert sub_dir_name in [directory.name for directory in directories]

    # THEN no files are returned
    assert some_file not in [directory.name for directory in directories]


def test_remove_directory_and_contents(path_with_directories_and_a_file: Path, some_file: str):
    """Test to remove a directory and all its contents."""

    # GIVEN a path to a directory
    assert path_with_directories_and_a_file.exists()

    # WHEN removing the directory
    remove_directory_and_contents(path_with_directories_and_a_file)

    # THEN the directory and its contents should no longer exist
    assert not path_with_directories_and_a_file.exists()
    assert not Path(path_with_directories_and_a_file, some_file).exists()



from pathlib import Path
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.models.slurm.sbatch import Sbatch
from cg.services.slurm_service.slurm_service import SlurmService


class SlurmCLIService(SlurmService):

    def __init__(self) -> None:
        self.client = SlurmAPI()

    def submit_job(self, job_config: Sbatch) -> int:
        sbatch_content: str = self.client.generate_sbatch_content(job_config)
        sbatch_path: Path = self._get_sbatch_path(job_config)
        return self.client.submit_sbatch(sbatch_content=sbatch_content, sbatch_path=sbatch_path)

    def _get_sbatch_path(self, job_config: Sbatch) -> Path:
        return Path(job_config.log_dir, f"{job_config.job_name}.sh")



from pathlib import Path

from click.testing import CliRunner
from mock import mock

from cg.cli.workflow.raredisease.base import managed_variants, panel
from cg.constants.scout import ScoutExportFileName
from cg.io.txt import read_txt
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.conftest import create_process_response

SUBPROCESS_RUN_FUNCTION_NAME: str = "cg.utils.commands.subprocess.run"


def test_panel_dry_run(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_panel_output: str,
):
    # GIVEN a case

    # GIVEN that, the Scout command writes the panel to stdout
    with mock.patch(
        SUBPROCESS_RUN_FUNCTION_NAME,
        return_value=create_process_response(std_out=scout_panel_output),
    ):
        # WHEN creating a panel file using dry-run
        result = cli_runner.invoke(
            panel, [raredisease_case_id, "--dry-run"], obj=raredisease_context
        )

    # THEN the output should contain the output from Scout
    assert result.stdout.strip() == scout_panel_output


def test_panel_file_is_written(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_panel_output: str,
):
    # GIVEN an analysis API
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # GIVEN a case

    # GIVEN that, the Scout command writes the panel to stdout
    with mock.patch(
        SUBPROCESS_RUN_FUNCTION_NAME,
        return_value=create_process_response(std_out=scout_panel_output),
    ):
        # WHEN creating a panel file
        cli_runner.invoke(panel, [raredisease_case_id], obj=raredisease_context)

    panel_file = Path(analysis_api.root, raredisease_case_id, ScoutExportFileName.PANELS)

    # THEN the file should exist
    assert panel_file.exists()

    # THEN the file should contain the output from Scout
    file_content: str = read_txt(file_path=panel_file, read_to_string=True)
    assert file_content == scout_panel_output


def test_managed_variants_is_written(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_export_manged_variants_output: str,
):
    # GIVEN an analysis API
    analysis_api: RarediseaseAnalysisAPI = raredisease_context.meta_apis["analysis_api"]

    # GIVEN a case

    # GIVEN that, the Scout command writes the managed variants to stdout
    with mock.patch(
        SUBPROCESS_RUN_FUNCTION_NAME,
        return_value=create_process_response(std_out=scout_export_manged_variants_output),
    ):
        # WHEN creating a managed_variants file
        cli_runner.invoke(managed_variants, [raredisease_case_id], obj=raredisease_context)

    managed_variants_file = Path(
        analysis_api.root, raredisease_case_id, ScoutExportFileName.MANAGED_VARIANTS
    )

    # THEN the file should exist
    assert managed_variants_file.exists()

    # THEN the file should contain the output from Scout
    file_content: str = read_txt(file_path=managed_variants_file, read_to_string=True)
    assert file_content == scout_export_manged_variants_output


def test_managed_variants_dry_run(
    raredisease_case_id: str,
    cli_runner: CliRunner,
    raredisease_context: CGConfig,
    scout_export_manged_variants_output: str,
):
    # GIVEN a case

    # GIVEN that, the Scout command writes the managed variants to stdout
    with mock.patch(
        SUBPROCESS_RUN_FUNCTION_NAME,
        return_value=create_process_response(std_out=scout_export_manged_variants_output),
    ):
        # WHEN creating a managed_variants file using dry run
        result = cli_runner.invoke(
            managed_variants, [raredisease_case_id, "--dry-run"], obj=raredisease_context
        )

    # THEN the result should contain the output from Scout
    assert result.stdout.strip() == scout_export_manged_variants_output



from pathlib import Path

import pytest

from cg.models.cg_config import CGConfig
from cg.services.validate_file_transfer_service.validate_file_transfer_service import (
    ValidateFileTransferService,
)
from cg.services.validate_file_transfer_service.validate_pacbio_file_transfer_service import (
    ValidatePacbioFileTransferService,
)


@pytest.fixture
def manifest_file(fixtures_dir: Path) -> Path:
    """Return the path to a manifest file."""
    return Path(fixtures_dir, "services", "validate_file_transfer_service", "pacbio_transfer_done")


@pytest.fixture
def manifest_file_fail(fixtures_dir: Path) -> Path:
    """Return the path to a manifest file."""
    return Path(
        fixtures_dir, "services", "validate_file_transfer_service", "pacbio_transfer_done_fail"
    )


@pytest.fixture
def transfer_source_dir(tmp_path: Path) -> Path:
    """Return the path to a source directory with files in different directories."""
    tmp_path.mkdir(exist_ok=True)
    tmp_path.joinpath("tree_1").mkdir()
    tmp_path.joinpath("tree_1/tree2").mkdir()
    tmp_path.joinpath("tree_1/tree2/tree3").mkdir()
    tmp_path.joinpath("file1").touch()
    tmp_path.joinpath("tree_1/file2").touch()
    tmp_path.joinpath("tree_1/file3").touch()
    tmp_path.joinpath("tree_1/tree2/file4").touch()
    tmp_path.joinpath("tree_1/tree2/tree3/file5").touch()
    return tmp_path


@pytest.fixture
def validate_file_transfer_service() -> ValidateFileTransferService:
    """Return the validate file transfer service."""
    return ValidateFileTransferService()


@pytest.fixture
def expected_file_names_in_manifest() -> list[str]:
    file_names: list[str] = []
    for i in range(1, 6):
        file_names.append(f"file{i}")
    return file_names


@pytest.fixture
def pacbio_run_id() -> str:
    """Return a PacBio run ID."""
    return "r1123_221421_12321"


@pytest.fixture
def smrt_cell_id() -> str:
    """Return a PacBio smrt cell ID."""
    return "1_A1_0"


@pytest.fixture
def metadata_dir_name() -> str:
    """Return a metadata directory name."""
    return "metadata"


@pytest.fixture
def tmp_manifest_file_name(pacbio_run_id: str, smrt_cell_id: str) -> str:
    """Return a temporary manifest file name."""
    return f"{pacbio_run_id}_{smrt_cell_id}.transferdone"


@pytest.fixture
def pacbio_runs_dir(
    tmp_path: Path, pacbio_run_id: str, smrt_cell_id, metadata_dir_name: str, tmp_manifest_file_name
) -> Path:
    """Return the path to a directory with PacBio runs."""
    tmp_path.mkdir(exist_ok=True)
    tmp_path.joinpath(pacbio_run_id).mkdir()
    tmp_path.joinpath(pacbio_run_id).joinpath(smrt_cell_id).mkdir()
    tmp_path.joinpath(pacbio_run_id).joinpath(smrt_cell_id).joinpath(metadata_dir_name).mkdir()
    tmp_path.joinpath(pacbio_run_id).joinpath(smrt_cell_id).joinpath(metadata_dir_name).joinpath(
        tmp_manifest_file_name
    ).touch()
    tmp_path.joinpath("not_this_dir").mkdir()
    return tmp_path


@pytest.fixture
def validate_pacbio_service(
    cg_context: CGConfig, pacbio_runs_dir: Path
) -> ValidatePacbioFileTransferService:
    """Return a PacBio file transfer service."""
    validate_pacbio_file_transfer_service = ValidatePacbioFileTransferService(config=cg_context)
    validate_pacbio_file_transfer_service.data_dir = pacbio_runs_dir.as_posix()
    validate_pacbio_file_transfer_service.trigger_dir = pacbio_runs_dir.as_posix()
    return validate_pacbio_file_transfer_service



from datetime import datetime
from pathlib import Path


from cg.constants.demultiplexing import UNDETERMINED
from cg.constants.devices import DeviceType
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.services.bcl_convert_metrics_service.parser import MetricsParser
from cg.store.models import (
    SampleLaneSequencingMetrics,
    IlluminaSampleSequencingMetrics,
)
from cg.store.store import Store
from cg.utils.flow_cell import get_flow_cell_id


class BCLConvertMetricsService:
    def create_sample_lane_sequencing_metrics_for_flow_cell(
        self,
        flow_cell_directory: Path,
    ) -> list[SampleLaneSequencingMetrics]:
        """Parse the demultiplexing metrics data into the sequencing statistics model."""
        metrics_parser = MetricsParser(flow_cell_directory)
        sample_internal_ids: list[str] = metrics_parser.get_sample_internal_ids()
        sample_lane_sequencing_metrics: list[SampleLaneSequencingMetrics] = []

        for sample_internal_id in sample_internal_ids:
            for lane in metrics_parser.get_lanes_for_sample(sample_internal_id=sample_internal_id):
                metrics: SampleLaneSequencingMetrics = self._create_bcl_convert_sequencing_metrics(
                    sample_internal_id=sample_internal_id, lane=lane, metrics_parser=metrics_parser
                )
                sample_lane_sequencing_metrics.append(metrics)
        return sample_lane_sequencing_metrics

    def create_undetermined_non_pooled_metrics(
        self,
        flow_cell: FlowCellDirectoryData,
    ) -> list[SampleLaneSequencingMetrics]:
        """Return sequencing metrics for any undetermined reads in non-pooled lanes."""

        non_pooled_lanes_and_samples: list[tuple[int, str]] = (
            flow_cell.sample_sheet.get_non_pooled_lanes_and_samples()
        )
        metrics_parser = MetricsParser(flow_cell.path)
        undetermined_metrics: list[SampleLaneSequencingMetrics] = []

        for lane, sample_internal_id in non_pooled_lanes_and_samples:
            if not metrics_parser.has_undetermined_reads_in_lane(lane):
                continue

            # Passing Undetermined as the sample id is required to extract the undetermined reads data.
            # BclConvert tags undetermined reads in a lane with the sample id "Undetermined".
            metrics: SampleLaneSequencingMetrics = self._create_bcl_convert_sequencing_metrics(
                sample_internal_id=UNDETERMINED, lane=lane, metrics_parser=metrics_parser
            )
            metrics.sample_internal_id = sample_internal_id
            undetermined_metrics.append(metrics)
        return undetermined_metrics

    @staticmethod
    def _create_bcl_convert_sequencing_metrics(
        sample_internal_id: str, lane: int, metrics_parser: MetricsParser
    ) -> SampleLaneSequencingMetrics:
        """Create sequencing metrics for a sample in a lane."""
        flow_cell_id: str = get_flow_cell_id(metrics_parser.bcl_convert_demultiplex_dir.name)

        total_reads: int = metrics_parser.calculate_total_reads_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        q30_bases_percent: float = metrics_parser.get_q30_bases_percent_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        mean_quality_score: float = metrics_parser.get_mean_quality_score_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        return SampleLaneSequencingMetrics(
            sample_internal_id=sample_internal_id,
            flow_cell_name=flow_cell_id,
            flow_cell_lane_number=lane,
            sample_total_reads_in_lane=total_reads,
            sample_base_percentage_passing_q30=q30_bases_percent,
            sample_base_mean_quality_score=mean_quality_score,
            created_at=datetime.now(),
        )

    @staticmethod
    def create_sample_run_metrics(
        sample_internal_id: str,
        lane: int,
        metrics_parser: MetricsParser,
        instrument_run_id: int,
        store: Store,
    ) -> IlluminaSampleSequencingMetrics:
        """Create sequencing metrics for all lanes in a flow cell."""

        total_reads: int = metrics_parser.calculate_total_reads_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        q30_bases_percent: float = metrics_parser.get_q30_bases_percent_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        mean_quality_score: float = metrics_parser.get_mean_quality_score_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        sample_id: int = store.get_sample_by_internal_id(sample_internal_id).id

        yield_: float = metrics_parser.get_yield_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        yield_q30: float = metrics_parser.get_yield_q30_for_sample_in_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )

        return IlluminaSampleSequencingMetrics(
            sample_id=sample_id,
            instrument_run_id=instrument_run_id,
            type=DeviceType.ILLUMINA,
            flow_cell_lane=lane,
            total_reads_in_lane=total_reads,
            base_percentage_passing_q30=q30_bases_percent,
            base_mean_quality_score=mean_quality_score,
            yield_=yield_,
            yield_q30=yield_q30,
            created_at=datetime.now(),
        )

    def create_sample_sequencing_metrics_for_flow_cell(
        self,
        flow_cell_directory: Path,
        instrument_run_id: int,
        store: Store,
    ) -> list[IlluminaSampleSequencingMetrics]:
        """Parse the demultiplexing metrics data into the sequencing statistics model."""
        metrics_parser = MetricsParser(flow_cell_directory)
        sample_internal_ids: list[str] = metrics_parser.get_sample_internal_ids()
        sample_lane_sequencing_metrics: list[IlluminaSampleSequencingMetrics] = []

        for sample_internal_id in sample_internal_ids:
            for lane in metrics_parser.get_lanes_for_sample(sample_internal_id=sample_internal_id):
                metrics: IlluminaSampleSequencingMetrics = self.create_sample_run_metrics(
                    sample_internal_id=sample_internal_id,
                    lane=lane,
                    metrics_parser=metrics_parser,
                    instrument_run_id=instrument_run_id,
                    store=store,
                )
                sample_lane_sequencing_metrics.append(metrics)
        return sample_lane_sequencing_metrics



"""Module with validator classes for the sample sheet."""

import logging
from pathlib import Path

from pydantic import ValidationError

from cg.apps.demultiplex.sample_sheet.override_cycles_validator import OverrideCyclesValidator
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import (
    get_flow_cell_samples_from_content,
    get_raw_samples_from_content,
    validate_samples_unique_per_lane,
)
from cg.apps.demultiplex.sample_sheet.sample_models import FlowCellSample
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import NAME_TO_INDEX_SETTINGS, SampleSheetBCLConvertSections
from cg.exc import OverrideCyclesError, SampleSheetError
from cg.io.controller import ReadFile

LOG = logging.getLogger(__name__)


class SampleSheetValidator:
    """Class for validating the content of a sample sheet."""

    def __init__(self):
        """Instantiate the class with a sample sheet file path or sample sheet content."""
        self.content: list[list[str]] | None = None
        self.read1_cycles: int | None = None
        self.read2_cycles: int | None = None
        self.index1_cycles: int | None = None
        self.index2_cycles: int | None = None
        self.is_index2_reverse_complement: bool | None = None

    def set_sample_sheet_content(self, content: list[list[str]]) -> None:
        """Set the sample sheet content."""
        self.content = content

    def _validate_all_sections_present(self) -> None:
        """
        Returns whether the sample sheet has the four mandatory sections:
            - Header
            - Reads
            - BCLConvert Settings
            - BCLConvert Data
        Raises:
            SampleSheetError if the sample sheet does not have all the sections.
        """
        LOG.debug("Validating that the sample sheet has all the necessary sections")
        has_header: bool = [SampleSheetBCLConvertSections.Header.HEADER] in self.content
        has_cycles: bool = [SampleSheetBCLConvertSections.Reads.HEADER] in self.content
        has_settings: bool = [SampleSheetBCLConvertSections.Settings.HEADER] in self.content
        has_data: bool = [SampleSheetBCLConvertSections.Data.HEADER] in self.content
        if not all([has_header, has_cycles, has_settings, has_data]):
            message: str = "Sample sheet does not have all the necessary sections"
            LOG.error(message)
            raise SampleSheetError(message)

    def _get_index_settings_name(self) -> str:
        """
        Find the entry in the sample sheet holding the index settings name, which has the form:
        `IndexSettings,<index_setting_name>`, and extract its value.
        """

        for row in self.content:
            if SampleSheetBCLConvertSections.Header.INDEX_SETTINGS in row:
                LOG.debug(f"Found index settings: {row[1]}")
                return row[1]
        message: str = "No index settings found in sample sheet"
        LOG.error(message)
        raise SampleSheetError(message)

    def _set_is_index2_reverse_complement(self) -> None:
        """Return whether the index2 override cycles value is reverse-complemented."""
        LOG.debug("Looking for index settings in the sample sheet")
        settings_name: str = self._get_index_settings_name()
        self.is_index2_reverse_complement = NAME_TO_INDEX_SETTINGS[
            settings_name
        ].are_i5_override_cycles_reverse_complemented

    def _get_cycle(self, cycle_name: str, nullable: bool = False) -> int | None:
        """
        Return the cycle from the sample sheet given the cycle name. Set nullable to True to
        return None if the cycle is not found.
        Raises:
            SampleSheetError if the cycle is not found and nullable is False.
        """
        for row in self.content:
            if cycle_name in row:
                return int(row[1])
        if not nullable:
            message: str = f"No {cycle_name} found in sample sheet"
            LOG.error(message)
            raise SampleSheetError(message)

    def _set_cycles(self):
        """Set values to the run cycle attributes."""
        LOG.debug("Looking for read and index run cycles in the sample sheet")
        self.read1_cycles = self._get_cycle(SampleSheetBCLConvertSections.Reads.READ_CYCLES_1)
        self.read2_cycles = self._get_cycle(SampleSheetBCLConvertSections.Reads.READ_CYCLES_2)
        self.index1_cycles = self._get_cycle(SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_1)
        self.index2_cycles = self._get_cycle(
            cycle_name=SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2, nullable=True
        )

    def _validate_samples(self) -> None:
        """
        Determine if the samples have the correct attributes and are unique per lane.
        Raises:
            SampleSheetError if the samples do not have the correct attributes based on their model
            or are not unique per lane.
        """
        LOG.debug("Validating samples")
        try:
            validated_samples: list[FlowCellSample] = get_flow_cell_samples_from_content(
                sample_sheet_content=self.content
            )
        except ValidationError as error:
            LOG.error("Sample sheet failed validation: samples are not in the correct format")
            raise SampleSheetError from error
        validate_samples_unique_per_lane(validated_samples)

    def _validate_override_cycles(self) -> None:
        """Determine if the samples' override cycles are valid.
        Raises:
            SampleSheetError if any of the samples' override cycles are not valid.
        """
        LOG.debug("Validating override cycles for all samples")
        validator = OverrideCyclesValidator(
            run_read1_cycles=self.read1_cycles,
            run_read2_cycles=self.read2_cycles,
            run_index1_cycles=self.index1_cycles,
            run_index2_cycles=self.index2_cycles,
            is_reverse_complement=self.is_index2_reverse_complement,
        )
        samples: list[dict[str, str]] = get_raw_samples_from_content(self.content)
        for sample in samples:
            try:
                validator.validate_sample(sample)
            except OverrideCyclesError as error:
                raise SampleSheetError from error

    def validate_sample_sheet_from_content(self, content: list[list[str]]) -> None:
        """
        Determine if the BCLConvert sample sheet is valid, which means:
        - All sections are present
        - The index settings are specified in the sample sheet header
        - The read and index cycles are specified in the sample sheet's reads section
        - The samples have the correct attributes
        - The override cycles are valid
        Raises:
            SampleSheetError: If the sample sheet is not valid.
        """
        self.set_sample_sheet_content(content)
        LOG.debug("Validating sample sheet")
        self._validate_all_sections_present()
        self._set_is_index2_reverse_complement()
        self._set_cycles()
        self._validate_samples()
        self._validate_override_cycles()
        LOG.info("Samplesheet passed validation")

    def validate_sample_sheet_from_file(self, file_path: Path) -> None:
        """
        Validate a sample sheet given the path to the file.
        Raises:
            SampleSheetError: If the sample sheet is not valid.
        """
        content: list[list[str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.CSV, file_path=file_path
        )
        self.validate_sample_sheet_from_content(content)

    def get_sample_sheet_object_from_file(self, file_path: Path) -> SampleSheet:
        """Return a sample sheet object given the path to the file.
        Raises:
            SampleSheetError: If the sample sheet is not valid.
        """
        self.validate_sample_sheet_from_file(file_path)
        samples: list[FlowCellSample] = get_flow_cell_samples_from_content(self.content)
        return SampleSheet(samples=samples)



"""Tests for the sequencing metrics parser API."""

from pathlib import Path

from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.services.bcl_convert_metrics_service.bcl_convert_metrics_service import (
    BCLConvertMetricsService,
)

from cg.services.bcl_convert_metrics_service.parser import MetricsParser
from cg.store.models import SampleLaneSequencingMetrics


def test_create_sample_lane_sequencing_metrics_for_flow_cell(
    bcl_convert_metrics_dir_path: Path,
    parsed_bcl_convert_metrics: MetricsParser,
    bcl_convert_metrics_service: BCLConvertMetricsService,
):
    """Test to create sequencing statistics from bcl convert metrics."""
    # GIVEN a parsed bcl convert metrics file

    # WHEN creating sequencing statistics from bcl convert metrics
    sequencing_statistics_list: list[SampleLaneSequencingMetrics] = (
        bcl_convert_metrics_service.create_sample_lane_sequencing_metrics_for_flow_cell(
            flow_cell_directory=bcl_convert_metrics_dir_path,
        )
    )

    # THEN assert that Sequencing statistics are created
    for sequencing_statistics in sequencing_statistics_list:
        assert isinstance(sequencing_statistics, SampleLaneSequencingMetrics)

    # THEN assert that the number of sequencing statistics created is correct
    assert (
        len(sequencing_statistics_list)
        == len(parsed_bcl_convert_metrics.get_sample_internal_ids()) * 2
    )


def test_create_undetermined_non_pooled_metrics(
    hiseq_x_single_index_demultiplexed_flow_cell_with_sample_sheet: FlowCellDirectoryData,
    bcl_convert_metrics_service: BCLConvertMetricsService,
):
    """Test creating undetermined sequencing statistics from demultiplex metrics."""
    # GIVEN a directory with a demultiplexed flow cell with undetermined reads

    # WHEN creating undetermined sequencing statistics from bcl convert metrics
    metrics: list[SampleLaneSequencingMetrics] = (
        bcl_convert_metrics_service.create_undetermined_non_pooled_metrics(
            flow_cell=hiseq_x_single_index_demultiplexed_flow_cell_with_sample_sheet
        )
    )

    # THEN metrics are created for the undetermined reads
    assert isinstance(metrics, list)
    assert isinstance(metrics[0], SampleLaneSequencingMetrics)


def test_create_undetermined_non_pooled_metrics_for_existing_lane_without_undetermined_reads(
    bcl_convert_metrics_dir_path: Path,
    bcl_convert_metrics_service: BCLConvertMetricsService,
):
    """
    Test creating undetermined sequencing statistics from demultiplex metrics without undetermined
    reads.
    """

    # GIVEN a directory with a demultiplexed flow cell without undetermined reads in a lane
    flow_cell = FlowCellDirectoryData(bcl_convert_metrics_dir_path)
    sample_sheet_path = Path(bcl_convert_metrics_dir_path, "SampleSheet.csv")
    flow_cell.set_sample_sheet_path_hk(hk_path=sample_sheet_path)

    # WHEN creating undetermined sequencing statistics specifying a lane without undetermined reads
    metrics: list[SampleLaneSequencingMetrics] = (
        bcl_convert_metrics_service.create_undetermined_non_pooled_metrics(flow_cell=flow_cell)
    )

    # THEN an empty list is returned
    assert not metrics
